"""
Unit tests for the RedshiftCatalog implementation
"""

import unittest
import numpy as np

from nessie.catalog import RedshiftCatalog
from nessie.cosmology import FlatCosmology
from nessie.helper_funcs import create_density_function


class TestRedshiftCatalog(unittest.TestCase):
    """
    Testing the RedshiftCatalog class and methods in the catalog module.
    """

    def test_it_runs(self):
        """
        Simple tets to ensure basic funtionality.
        """
        ra_array = np.array([120.0, 120.0, 50.0])
        dec_array = np.array([-34.0, -34.0, 23.0])
        redshift_array = np.array([0.2, 0.2, 0.6])
        velocity_errors = np.array([50.0, 50.0, 50.0])
        absolute_magnitudes = np.array([-18.0, -18.0, -18.0])

        cosmo = FlatCosmology(0.7, 0.3)
        completeness = np.repeat(0.98, len(ra_array))
        density_function = lambda z: np.repeat(0.2, len(z))
        cat = RedshiftCatalog(
            ra_array, dec_array, redshift_array, density_function, cosmo
        )
        cat.set_completeness(completeness)

        group = cat.get_raw_groups(0.06, 18)
        cat.run_fof(0.06, 18)
        group_catalog = cat.calculate_group_table(absolute_magnitudes, velocity_errors)

        np.testing.assert_array_equal(group["galaxy_id"], np.array([1, 2]))
        np.testing.assert_array_equal(group["group_id"], np.array([1, 1]))
        np.testing.assert_array_equal(cat.group_ids, np.array([1, 1, -1]))
        self.assertEqual(round(group_catalog["iter_ra"][0]), 120.0)
        self.assertEqual(round(group_catalog["iter_dec"][0]), -34.0)
        self.assertEqual(group_catalog["multiplicity"][0], 2)

    def test_completeness_is_automatically_set(self):
        """
        Checking that the completeness is updated appropriately
        """
        ra_array = np.array([120.0, 120.0, 50.0])
        dec_array = np.array([-34.0, -34.0, 23.0])
        redshift_array = np.array([0.2, 0.2, 0.6])
        cosmo = FlatCosmology(0.7, 0.3)
        density_function = lambda z: np.repeat(0.2, len(z))
        cat = RedshiftCatalog(
            ra_array, dec_array, redshift_array, density_function, cosmo
        )
        cat.set_completeness()
        np.testing.assert_array_equal(cat.completeness, np.ones(len(ra_array)))

    def test_completeness_can_be_calculated(self):
        """
        Checking that we can calculate the completeness from a target catalog.
        """
        ra_array = np.array([120.0, 120.0])
        dec_array = np.array([-34.0, -34.0])
        target_ra = np.array([120.0, 120.0, 300.0])
        target_dec = np.array([-34.0, -34.0, -34.0])
        redshift_array = np.array([0.2, 0.2])
        cosmo = FlatCosmology(0.7, 0.3)
        density_function = lambda z: np.repeat(0.2, len(z))
        cat = RedshiftCatalog(
            ra_array, dec_array, redshift_array, density_function, cosmo
        )
        cat.calculate_completeness(target_ra, target_dec, radii=np.array([0.1, 0.1]))
        np.testing.assert_equal(cat.completeness, np.ones(2))

        cat.ra_array = np.array([120.0])
        cat.dec_array = np.array([-34.0])
        cat.calculate_completeness(target_ra, target_dec, radii=np.array([0.1]))
        np.testing.assert_equal(cat.completeness, np.array([0.5]))

    def test_getting_group_ids_works_on_simple_case(self):
        """
        More complete example. Especially to test the 1-indexing between R and python.
        """
        b0 = 100.0
        r0 = 180.0
        ras = np.array(
            [20.0, 20.0, 20.0, 140.0, 140.0, 140.0, 100.0, 100.0, 0.0, 180.0]
        )
        decs = np.array([-50.0, -50.0, -50.0, 0.0, 0.0, 0.0, 90.0, 90.0, 45.0, -45.0])
        redshifts = np.array([0.2, 0.2, 0.2, 0.3, 0.3, 0.3, 0.4, 0.4, 0.2, 0.2])
        cosmo = FlatCosmology(0.7, 0.3)

        random_redshifts = np.random.normal(0.2, 0.1, 20000)
        random_redshifts = random_redshifts[random_redshifts > 0]
        rho_mean = create_density_function(
            random_redshifts, 20000, 0.001, cosmology=cosmo
        )

        cat = RedshiftCatalog(ras, decs, redshifts, rho_mean, cosmo)
        cat.set_completeness()
        cat.run_fof(b0, r0)

        expected = np.array([1, 1, 1, 2, 2, 2, 3, 3, -1, -1])
        np.testing.assert_array_equal(cat.group_ids, expected)

    def test_comparison_to_mocks_is_working(self):
        """
        Testing that class implementation of the s-score is working.
        """
        group_ids = np.array([1, 1, 1, 2, 2, 2, 2, 3, 3, -1, -1])
        mock_group_ids = np.array([100, 100, 100, 200, 200, -1, -1, -1, -1, -1, -1])
        ras = np.full(11, 120.0)
        decs = np.full(11, 45.0)
        redshifts = np.full(11, 0.2)
        cosmo = FlatCosmology(0.7, 0.3)
        rho_mean = lambda z: np.ones(len(z))

        cat = RedshiftCatalog(ras, decs, redshifts, rho_mean, cosmo)
        cat.set_completeness()
        cat.group_ids = group_ids
        cat.mock_group_ids = mock_group_ids

        metrics = cat.compare_to_mock()
        self.assertAlmostEqual(metrics, 0.1111, places=3)


if __name__ == "__main__":
    unittest.main()



"""
Unit test checking the validation functionality in Nessie.
"""

import warnings
import unittest
import numpy as np
from nessie.helper_funcs import validate, ValidationType


class TestValidateFunction(unittest.TestCase):
    """Unit tests for the validate function."""

    # --- RA ---
    def test_ra_valid(self):
        """Check that RA values in valid range pass without error."""
        try:
            validate(np.array([0.0, 180.0, 359.9]), ValidationType.RA)
        except Exception as e:
            self.fail(f"RA validation raised unexpectedly: {e}")

    def test_ra_below_zero_raises(self):
        """Check that negative RA values raise an error."""
        with self.assertRaises(ValueError):
            validate(np.array([-1.0]), ValidationType.RA)

    def test_ra_above_360_raises(self):
        """Check that RA values above 360 raise an error."""
        with self.assertRaises(ValueError):
            validate(np.array([361.0]), ValidationType.RA)

    # --- DEC ---
    def test_dec_valid(self):
        """Check that Declination values in valid range pass without error."""
        try:
            validate(np.array([-90.0, 0.0, 89.9]), ValidationType.DEC)
        except Exception as e:
            self.fail(f"Dec validation raised unexpectedly: {e}")

    def test_dec_below_negative_90_raises(self):
        """Check that Declination values below -90 raise an error."""
        with self.assertRaises(ValueError):
            validate(np.array([-91.0]), ValidationType.DEC)

    def test_dec_above_90_raises(self):
        """Check that Declination values above 90 raise an error."""
        with self.assertRaises(ValueError):
            validate(np.array([91.0]), ValidationType.DEC)

    # --- Redshift ---
    def test_redshift_valid(self):
        """Check that redshift values in valid range pass without error."""
        try:
            validate(np.array([0.0, 1.0, 10.0]), ValidationType.REDSHIFT)
        except Exception as e:
            self.fail(f"Redshift validation raised unexpectedly: {e}")

    def test_redshift_negative_raises(self):
        """Check that negative redshift values raise an error."""
        with self.assertRaises(ValueError):
            validate(np.array([-0.1]), ValidationType.REDSHIFT)

    def test_redshift_large_warns(self):
        """Check that redshift values above 1100 raise a warning."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            validate(np.array([1200.0]), ValidationType.REDSHIFT)
            self.assertTrue(
                any("redshifts are very large" in str(warn.message) for warn in w)
            )

    # --- Absolute Magnitude ---
    def test_abs_mag_valid(self):
        """Check that valid absolute magnitudes pass without error."""
        try:
            validate(np.array([-21.0, -10.0]), ValidationType.ABS_MAG)
        except Exception as e:
            self.fail(f"Absolute magnitude validation raised unexpectedly: {e}")

    def test_abs_mag_too_faint_warns(self):
        """Check that absolute magnitudes brighter than -4 raise a warning."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            validate(np.array([-3.0]), ValidationType.ABS_MAG)
            self.assertTrue(
                any(
                    "absolute magnitudes look unusual" in str(warn.message)
                    for warn in w
                )
            )

    def test_abs_mag_too_bright_warns(self):
        """Check that absolute magnitudes fainter than -50 raise a warning."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            validate(np.array([-51.0]), ValidationType.ABS_MAG)
            self.assertTrue(
                any(
                    "absolute magnitudes look unusual" in str(warn.message)
                    for warn in w
                )
            )

    # --- Completeness ---
    def test_completeness_valid(self):
        """Check that completeness between 0 and 1 passes without error."""
        try:
            validate(np.array([0.85, 0.1]), ValidationType.COMPLETENESS)
        except Exception as e:
            self.fail(f"Completeness validation raised unexpectedly: {e}")

    def test_completeness_not_numeric_raises(self):
        """Check that non-numeric completeness raises a TypeError."""
        with self.assertRaises(TypeError):
            validate(np.array(["not a number", 0.1]), ValidationType.COMPLETENESS)

    def test_completeness_below_zero_raises(self):
        """Check that completeness below 0 raises a ValueError."""
        with self.assertRaises(ValueError):
            validate(np.array([-0.1, 0.1]), ValidationType.COMPLETENESS)

    def test_completeness_above_one_raises(self):
        """Check that completeness above 1 raises a ValueError."""
        with self.assertRaises(ValueError):
            validate(np.array([1.1, 0.1]), ValidationType.COMPLETENESS)

    def test_vel_errors_valid(self):
        """Check that the velocity errors above zero pass without error."""
        try:
            validate(np.array([50.0, 50.0]), ValidationType.VEL_ERR)
        except Exception as e:
            self.fail(f"Velocity Errors failed unexpectedly: {e}")

    def test_vel_errors_warning(self):
        """Checking that there is a warning when the velocity errors seem too large."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            validate(np.array([2001]), ValidationType.VEL_ERR)
            self.assertTrue(
                any(
                    "Warning: velocity errors seem very large. Are units correct?"
                    in str(warn.message)
                    for warn in w
                )
            )

    def test_vel_errors_below_zero_raises(self):
        """Check that vel_err below 0 raises a ValueError."""
        with self.assertRaises(ValueError):
            validate(np.array([-50, 50]), ValidationType.VEL_ERR)

    # --- B0 ---
    def test_b0_valid(self):
        """Check that valid B0 value passes."""
        try:
            validate(1.5, ValidationType.B0)
        except Exception as e:
            self.fail(f"B0 validation raised unexpectedly: {e}")

    def test_b0_not_numeric_raises(self):
        """Check that non-numeric B0 raises a TypeError."""
        with self.assertRaises(TypeError):
            validate("bad", ValidationType.B0)

    def test_b0_negative_raises(self):
        """Check that negative B0 raises a ValueError."""
        with self.assertRaises(ValueError):
            validate(-1.0, ValidationType.B0)

    # --- R0 ---
    def test_r0_valid(self):
        """Check that valid R0 value passes."""
        try:
            validate(4.2, ValidationType.R0)
        except Exception as e:
            self.fail(f"R0 validation raised unexpectedly: {e}")

    def test_r0_not_numeric_raises(self):
        """Check that non-numeric R0 raises a TypeError."""
        with self.assertRaises(TypeError):
            validate(None, ValidationType.R0)

    def test_r0_negative_raises(self):
        """Check that negative R0 raises a ValueError."""
        with self.assertRaises(ValueError):
            validate(-1.0, ValidationType.R0)

    # --- General Array Validation ---
    def test_nan_in_array_raises(self):
        """Check that arrays containing NaN raise a ValueError."""
        with self.assertRaises(ValueError):
            validate(np.array([1.0, np.nan]), ValidationType.RA)

    def test_inf_in_array_raises(self):
        """Check that arrays containing Inf raise a ValueError."""
        with self.assertRaises(ValueError):
            validate(np.array([1.0, np.inf]), ValidationType.DEC)

    def test_non_numeric_array_raises(self):
        """Check that arrays with non-numeric values raise a TypeError."""
        with self.assertRaises(TypeError):
            validate(np.array(["a", "b"]), ValidationType.REDSHIFT)


if __name__ == "__main__":
    unittest.main()



"""
The Core redshift survey class which handles most the linking assignments and groups.
"""

from typing import Callable
import warnings
import numpy as np

from nessie_py import create_group_catalog, create_pair_catalog, calc_completeness_rust
from .cosmology import FlatCosmology
from .core_funcs import _find_groups
from .helper_funcs import calculate_s_total, validate, ValidationType, remap_ids


class RedshiftCatalog:
    """
    Stores the Redshift Catalog data that is needed for running the group finder.
    """

    def __init__(
        self,
        ra_array: np.ndarray[float],
        dec_array: np.ndarray[float],
        redshift_array: np.ndarray[float],
        density_function: Callable,
        cosmology: FlatCosmology,
    ) -> None:
        self.ra_array = ra_array
        self.dec_array = dec_array
        self.redshift_array = redshift_array
        self.density_function = density_function
        self.cosmology = cosmology

        self.completeness = None
        self.current_r0 = None
        self.current_b0 = None
        self.group_ids = None
        self.mock_group_ids = None

        validate(self.ra_array, ValidationType.RA)
        validate(self.dec_array, ValidationType.DEC)
        validate(self.redshift_array, ValidationType.REDSHIFT)

        if len(self.ra_array) != len(self.dec_array):
            raise ValueError("ra_array and dec_array should be the same length.")
        if len(self.redshift_array) != len(self.dec_array):
            raise ValueError(
                "redshift_array should be equal in length to the ra and dec arrays."
            )

    def calculate_completeness(
        self,
        ra_target: np.ndarray[float],
        dec_target: np.ndarray[float],
        radii: np.ndarray[float],
    ) -> None:
        """
        Calculate the completeness around the galaxy positions.

        Completeness is defined as the ratio of observed to targeted galaxies within a given
        angular radius.

        Parameters
        ----------
        ra_target : array_like
            Right Ascension (in degrees) of galaxies that were targeted for observation.

        dec_target : array_like
            Declination (in degrees) of galaxies that were targeted for observation.

        Returns
        -------
        completeness : ndarray
            Array of completeness values (floats between 0 and 1), one for each evaluation position.
        """
        validate(ra_target, ValidationType.RA)
        validate(dec_target, ValidationType.DEC)
        validate(radii, ValidationType.ANGLE)

        if len(ra_target) != len(dec_target):
            raise ValueError("ra_target and dec_target need to be the same length.")
        if len(ra_target) < len(self.ra_array):
            raise ValueError("target arrays must be larger than observed arrays!")
        if len(radii) != len(self.ra_array):
            raise ValueError(
                "radii should be the same length as the observed ra_array."
            )

        self.completeness = np.array(
            calc_completeness_rust(
                self.ra_array, self.dec_array, ra_target, dec_target, radii
            )
        )

    def set_completeness(self, completeness: np.ndarray[float] = None) -> None:
        """
        Sets the completeness of the redshift catalog to the given completeness array.
        If no arguments are given 100% completeness is assumed.
        """

        if completeness is None:
            completeness = np.ones(len(self.ra_array))
        if len(completeness) != len(self.ra_array):
            raise ValueError(
                "The completenes array must be equal to the number of galaxies in the survey."
            )
        validate(completeness, ValidationType.COMPLETENESS)
        self.completeness = np.array(completeness)

    def get_raw_groups(self, b0: float, r0: float, max_stellar_mass=1e15) -> dict:
        """
        Generate FoF links between galaxies based on spatial and redshift linking lengths. There is
        very little reason to have to run this yourself. In most cases it is more appropriate to run
        """
        co_dists = self.cosmology.comoving_distance(self.redshift_array)
        linking_lengths = np.array(
            self.density_function(self.redshift_array) ** (-1.0 / 3)
            * (self.completeness) ** (-1.0 / 3)
        )
        gal_rad = b0 * linking_lengths
        max_on_sky_radius = self.cosmology.virial_radius(
            max_stellar_mass, self.redshift_array
        )
        too_wide = np.where(gal_rad > max_on_sky_radius)[0]

        gal_rad[too_wide] = max_on_sky_radius[too_wide]
        linking_lengths_pos = gal_rad / (self.cosmology.h * co_dists)

        r_variable = (
            r0
            * (1 + self.redshift_array)
            / (
                np.sqrt(
                    self.cosmology.omega_m * (1 + self.redshift_array) ** 3
                    + self.cosmology.omega_lambda
                )
            )
        )
        linking_lengths_los = (gal_rad * r_variable) / self.cosmology.h
        max_los_distances = (
            self.cosmology.velocity_dispersion(max_stellar_mass, self.redshift_array)
            * (1 + self.redshift_array)
            / self.cosmology.h0_grow(self.redshift_array)
        )
        too_far = linking_lengths_los > max_los_distances
        linking_lengths_los[too_far] = max_los_distances[too_far]
        groups = _find_groups(
            self.ra_array,
            self.dec_array,
            co_dists,
            linking_lengths_pos,
            linking_lengths_los,
        )
        return groups

    def run_fof(self, b0: float, r0: float, max_stellar_mass=1e15) -> None:
        """
        Run the full Friends-of-Friends (FoF) algorithm and assign group IDs to all galaxies.
        Singleton galaxies (unlinked) are given group ID -1.
        """
        validate(b0, ValidationType.B0)
        validate(r0, ValidationType.R0)
        if self.completeness is None:
            raise ValueError(
                "No completeness array found. Run 'set_completeness' or 'calculate_completeness'."
            )

        group_links = self.get_raw_groups(b0, r0, max_stellar_mass)
        group_ids = np.ones(len(self.ra_array)) * -1

        group_ids[group_links["galaxy_id"] - 1] = group_links[
            "group_id"
        ]  # removing the 1-indexing
        self.group_ids = np.astype(group_ids, int)
        self.current_r0 = r0
        self.current_b0 = b0

    def calculate_group_table(
        self, absolute_magnitudes: np.ndarray[float], velocity_errors: np.ndarray[float]
    ) -> dict:
        """
        Generate a dictionary of group properties based on assigned group IDs.
        Must have run the group finder.
        """
        validate(absolute_magnitudes, ValidationType.ABS_MAG)
        validate(velocity_errors, ValidationType.VEL_ERR)

        if self.group_ids is None:
            raise InterruptedError(
                "Algorithm hasn't been run! Make sure to run_fof first."
            )
        group_cat = create_group_catalog(
            self.ra_array,
            self.dec_array,
            self.redshift_array,
            absolute_magnitudes,
            velocity_errors,
            self.group_ids,
            self.cosmology.omega_m,
            self.cosmology.omega_k,
            self.cosmology.omega_lambda,
            self.cosmology.hubble_constant,
        )
        return group_cat

    def calculate_pair_table(self, absolute_magnitudes: np.ndarray[float]) -> dict:
        """
        Generates a dictionary of pair properties based on assigned group IDs.
        Must have run the group finder.
        """
        validate(absolute_magnitudes, ValidationType.ABS_MAG)
        if self.group_ids is None:
            raise InterruptedError(
                "Algorithm hasn't been run! Make sure to run_fof first."
            )

        pair_cat = create_pair_catalog(
            self.ra_array,
            self.dec_array,
            self.redshift_array,
            absolute_magnitudes,
            self.group_ids,
        )
        return pair_cat

    def compare_to_mock(self, min_group_size=2):
        """
        Compares the current group_ids to a mock known grouping ids. Must have run the group finder
        and set both the mock_group_ids and singleton_id
        """
        if self.group_ids is None:
            raise InterruptedError(
                "No group ids found. Be sure to run the `run_fof` method"
            )

        if self.mock_group_ids is None:
            raise InterruptedError(
                "No mock group ids found. Be sure to set the mock groups ids."
            )
        elif -1 not in self.mock_group_ids:
            warnings.warn("No mock-groups found with -1 id. Are these set properly?")

        if self.completeness is None:
            raise ValueError(
                "No completeness array found. Run 'set_completeness' or 'calculate_completeness'."
            )

        safe_mock_ids = remap_ids(self.mock_group_ids)

        return calculate_s_total(self.group_ids, safe_mock_ids, min_group_size)



"""
Unit tests for the helper funcs module.
"""

import unittest
import numpy as np
import numpy.testing as npt

from nessie.helper_funcs import (
    create_density_function,
    calculate_s_total,
    remap_ids,
    gen_random_redshifts,
)
from nessie.cosmology import FlatCosmology


class TestDensityFunction(unittest.TestCase):
    """
    Testing the create_density_function in the helper_funcs module.
    """

    cosmo = FlatCosmology(1.0, 0.3)

    def test_against_nessie_r(self):
        """
        Testing against the implementation that is in the R nessie package.
        """
        redshifts = np.array([0.05, 0.1, 0.2, 0.3, 0.4])
        random_zs = np.loadtxt(
            "/Users/00115372/Desktop/GAMA_paper_plotter/gama_g09_randoms.txt",
            skiprows=1,
        )
        func = create_density_function(
            random_zs, len(random_zs) / 400, 0.001453924, self.cosmo
        )
        answers = np.array(
            [0.04775546, 0.024226080, 0.010309411, 0.003870724, 0.001073072]
        )
        result = func(redshifts)
        npt.assert_almost_equal(result, answers, decimal=2)


class TestCalculateSTotal(unittest.TestCase):
    """
    Testing the calculate_s_total function in helper functions.
    """

    def test_simple(self):
        """
        The R nessie package is already testing against the old algorithm. So we need only check
        that this score matches. And the rust code that it is based on is also tested. So we are
        really on testing that this runs.
        """

        measured_ids = np.array([0, 0, 0, 1, 1])
        mock_ids = np.array([0, 0, 0, -1, -1])
        result = calculate_s_total(measured_ids, mock_ids)
        self.assertEqual(result, 0.4)


class TestRemapId(unittest.TestCase):
    """Testing the remap_id function"""

    def test_string(self):
        """Testing that strings are remaped correctly."""
        bad = ["2mass1", "2mass2", -1, "2mass1", "2mass3"]
        good = [1, 2, -1, 1, 3]
        ans = remap_ids(bad)
        for r, a in zip(ans, good):
            self.assertEqual(r, a)

    def test_float(self):
        """Testing that floats work."""
        bad = [1.1, 2.2, -1, 1.1, 3.3]
        good = [1, 2, -1, 1, 3]
        ans = remap_ids(bad)
        for r, a in zip(ans, good):
            self.assertEqual(r, a)

    def test_long_int(self):
        """Testing that long ints work."""
        bad = [10**18, 10**18 + 1, -1, 10**18, 10**18 + 2]
        good = [1, 2, -1, 1, 3]
        ans = remap_ids(bad)
        for r, a in zip(ans, good):
            self.assertEqual(r, a)

    def test_shark(self):
        """Testing that works with shark ids."""
        shark_ids = [21826700000225, 21826700000235, -1, 21826700000225, 21826700000525]
        good = [1, 2, -1, 1, 3]
        ans = remap_ids(shark_ids)
        for r, a in zip(ans, good):
            self.assertEqual(r, a)


class TestGenRandoms(unittest.TestCase):
    """
    Testing that the gen_random_redshifts function.
    """

    def test(self):
        """testing basic run since this is a simple wrapper."""
        cosmo = FlatCosmology(1.0, 0.3)
        redshifts = np.random.normal(0.6, 0.3, 500)
        redshifts = redshifts[redshifts > 0]
        mags = np.random.random(len(redshifts)) * 4 + 15
        z_lim = np.max(redshifts) + 0.1
        maglim = 19.0
        zs = gen_random_redshifts(redshifts, mags, z_lim, maglim, cosmo, iterations=2, n_clone=100)
        self.assertEqual(round(len(zs) / (100*500)), 1)


if __name__ == "__main__":
    unittest.main()



"""
Cosmology module for handling the Flat LCDM cosmology calculations.
"""

import numpy as np
from nessie_py import (
    comoving_distances_at_z,
    distance_modulus,
    z_at_comoving_distances,
    calculate_max_rvirs,
    calculate_max_sigmas,
    h_at_z,
)


class FlatCosmology:
    """
    Core cosmology class
    """

    def __init__(self, h: float, omega_matter: float) -> None:
        """
        Read in h and Om0 and build cosmology from that.
        """
        self.h = h
        self.omega_m = omega_matter
        self.hubble_constant = 100 * self.h
        self.omega_lambda = 1 - self.omega_m
        self.omega_k = 0.0
        self.omgega_radiation = 0.0

    def comoving_distance(self, redshift: np.ndarray[float]) -> np.ndarray[float]:
        """
        Comoving distance for a vector of redshifts.
        """
        return np.array(
            comoving_distances_at_z(
                redshift,
                self.omega_m,
                self.omega_k,
                self.omega_lambda,
                self.hubble_constant,
            )
        )

    def dist_mod(self, redshift: np.ndarray[float]) -> np.ndarray[float]:
        """
        Distance modulus for an array of redshift values.
        """
        return np.array(
            distance_modulus(
                redshift,
                self.omega_m,
                self.omega_k,
                self.omega_lambda,
                self.hubble_constant,
            )
        )

    def z_at_comoving_distances(self, co_dist: np.ndarray[float]) -> np.ndarray[float]:
        """
        The inverse comoving distance function. Determines the redshift for the array of comoving
        distances.
        """
        return np.array(
            z_at_comoving_distances(
                co_dist,
                self.omega_m,
                self.omega_k,
                self.omega_lambda,
                self.hubble_constant,
            )
        )

    def virial_radius(self, solar_mass: float, redshifts: np.ndarray[float]):
        """
        Computes the virial radius of the halos with the given mass and redshift.
        """
        return np.array(
            calculate_max_rvirs(
                solar_mass,
                redshifts,
                self.omega_m,
                self.omega_k,
                self.omega_lambda,
                self.hubble_constant,
            )
        )

    def velocity_dispersion(self, solar_mass: float, redshifts: np.ndarray[float]):
        """
        Compute the velocity dispersion of halos with given mass and redshifts
        """
        return np.array(
            calculate_max_sigmas(
                solar_mass,
                redshifts,
                self.omega_m,
                self.omega_k,
                self.omega_lambda,
                self.hubble_constant,
            )
        )

    def h0_grow(self, redshifts: np.ndarray[float]) -> np.ndarray[float]:
        """
        Compute the redshift-dependent Hubble parameter H(z) in a flat cosmology
        """
        return np.array(
            h_at_z(
                redshifts,
                self.omega_m,
                self.omega_k,
                self.omega_lambda,
                self.hubble_constant,
            )
        )



"""
Module for quality of life helper functions which are not core to the algorithm.
"""

import warnings
from enum import Enum
from typing import Union

import numpy as np
from scipy.integrate import quad
from scipy.interpolate import interp1d
from statsmodels.nonparametric.kde import KDEUnivariate


from nessie_py import calculate_s_score, gen_randoms
from .cosmology import FlatCosmology


def gen_random_redshifts(
    redshifts: np.ndarray[float],
    mags: np.ndarray[float],
    z_lim: float,
    maglim: float,
    cosmo: FlatCosmology,
    n_clone: int = 400,
    iterations: int = 10,
) -> np.ndarray[float]:
    """
    Automatically models the n(z) using a randoms catalogue. This can be used to gen the density
    function.
    """

    return gen_randoms(
        redshifts,
        mags,
        z_lim,
        maglim,
        n_clone,
        iterations,
        cosmo.omega_m,
        cosmo.omega_k,
        cosmo.omega_lambda,
        cosmo.hubble_constant,
    )


def create_density_function(
    redshifts: np.ndarray,
    total_counts: int,
    survey_fractional_area: float,
    cosmology: FlatCosmology,
    binwidth: int = 40,
    interpolation_n: int = 10_000,
):
    """
    Running density function estimation (rho(z))

    Parameters
    ----------
    redshifts : array_like
        Redshift values used to estimate the redshift distribution.
    total_counts : float
        Total number of objects in the redshift survey.
    survey_fractional_area : float
        Fraction of the sky covered by the survey (relative to 4π steradians).
    cosmology : CustomCosmology
        A cosmology object with methods:
            - comoving_distance(z)
            - z_at_comoving_dist(d)
    binwidth : float
        Bin width in comoving Mpc (default = 40).
    interpolation_n : int
        Number of bins used for interpolation (default = 10,000).

    Returns
    -------
    rho_z_func : callable
        A function rho(z) that gives the running density at a given redshift.
    """
    comoving_distances = cosmology.comoving_distance(redshifts)

    kde = KDEUnivariate(comoving_distances)
    kde.fit(bw=binwidth, fft=True, gridsize=interpolation_n, cut=0)
    kde_x = kde.support
    kde_y = kde.density

    kde_func = interp1d(kde_x, kde_y, bounds_error=False, fill_value="extrapolate")

    # Running integral over each bin
    running_integral = np.array(
        [quad(kde_func, max(x - binwidth / 2, 0), x + binwidth / 2)[0] for x in kde_x]
    )

    # Running comoving volume per bin
    upper_volumes = (4 / 3) * np.pi * (kde_x + binwidth / 2) ** 3
    lower_volumes = (4 / 3) * np.pi * (kde_x - binwidth / 2) ** 3
    running_volume = survey_fractional_area * (upper_volumes - lower_volumes)

    # Convert comoving distance to redshift
    z_vals = cosmology.z_at_comoving_distances(kde_x)
    rho_vals = (total_counts * running_integral) / running_volume

    # Interpolate rho(z)
    rho_z_func = interp1d(
        z_vals, rho_vals, bounds_error=False, fill_value="extrapolate"
    )

    return rho_z_func


def calculate_s_total(
    measured_ids: np.ndarray[int],
    mock_group_ids: np.ndarray[int],
    min_group_size: int = 2,
) -> float:
    """
    Comparing measured groups to mock catalogue groups
    """
    return calculate_s_score(
        np.astype(measured_ids, int),
        np.astype(mock_group_ids, int),
        int(min_group_size),
    )


class ValidationType(Enum):
    """
    All the types of things we can validate
    """

    RA = "ra"
    DEC = "dec"
    REDSHIFT = "redshift"
    ABS_MAG = "absolute_mag"
    COMPLETENESS = "completeness"
    B0 = "b0"
    R0 = "r0"
    VEL_ERR = "vel_err"
    ANGLE = "angle"


def validate(value: np.ndarray[float] | float, valid_type: ValidationType) -> None:
    """
    Validates the given input which can be either an array or a float
    """

    # Checking arrays don't have non-numeric values
    if valid_type in {
        ValidationType.RA,
        ValidationType.DEC,
        ValidationType.ABS_MAG,
        ValidationType.REDSHIFT,
        ValidationType.COMPLETENESS,
        ValidationType.VEL_ERR,
        ValidationType.ANGLE,
    }:
        value = np.asarray(value)
        if not np.issubdtype(value.dtype, np.number):
            raise TypeError(f"{valid_type.value} must be numeric.")
        if np.isnan(value).any():
            raise ValueError(f"{valid_type.value} contains NaNs.")
        if np.isinf(value).any():
            raise ValueError(f"{valid_type.value} contains infinite values.")

    # Match-case for type-specific logic
    match valid_type:
        case ValidationType.RA:
            if np.any((value < 0) | (value > 360)):
                raise ValueError("RA values must be between 0 and 360.")

        case ValidationType.DEC:
            if np.any((value < -90) | (value > 90)):
                raise ValueError("Dec values must be between -90 and 90.")

        case ValidationType.REDSHIFT:
            if np.any(value < 0):
                raise ValueError("Redshifts cannot be negative.")
            if np.any(value > 1100):
                warnings.warn("Warning: redshifts are very large!")

        case ValidationType.ABS_MAG:
            if np.any((value > -4) | (value < -50)):
                warnings.warn("Warning: absolute magnitudes look unusual.")

        case ValidationType.COMPLETENESS:
            if np.any((value < 0) | (value > 1)):
                raise ValueError("Completeness must be between 0 and 1.")

        case ValidationType.VEL_ERR:
            if np.any(value < 0):
                raise ValueError("Velocity errors must be greater than 0.")
            if np.any(value > 2000):
                warnings.warn(
                    "Warning: velocity errors seem very large. Are units correct?"
                )

        case ValidationType.B0 | ValidationType.R0:
            if not isinstance(value, (float, int)):
                raise TypeError(f"{valid_type.value} must be a scalar number.")
            if value < 0:
                raise ValueError(f"{valid_type.value} cannot be negative.")

        case ValidationType.ANGLE:
            if np.any((value < 0) | (value > 360)):
                raise ValueError("Angle must be in degrees between 0 and 360.")

        case _:
            raise ValueError(
                f"Unknown property type: {valid_type.value}. Likely Enum needs updating."
            )


def remap_ids(id_array: np.ndarray[Union[int, str, float]]) -> np.ndarray[int]:
    """
    Renames the ids which could be in string, float, or long int into a safe int32 format.
    """
    unique_vals = np.unique(id_array)
    unique_vals = unique_vals[unique_vals != -1]
    unique_vals = unique_vals[unique_vals != "-1"]  # for string cases
    mapping = {val: i + 1 for i, val in enumerate(unique_vals)}  # plus to remove 0
    mapping[-1] = -1
    return np.vectorize(mapping.get)(id_array)



"""
Module to handle all the optimization code.
"""

from scipy.optimize import fmin
from .catalog import RedshiftCatalog


def optimize_nm(
    redshift_cat: RedshiftCatalog,
    min_group_size: int,
    b0_guess: float = 0.05,
    r0_guess: float = 30.0,
    max_stellar_mass: float = 1e15,
) -> tuple[float, float]:
    """
    Optimizes the b0 and r0 parameters using a Nelder-Mead simplex optimization approach to
    maximize the s_tot output for a given RedshiftCatalog object.

    Parameters
    ----------
    redshift_cat : RedshiftCatalog
        RedshiftCatalog object for which to optimize the grouping parameters.
    min_group_size : int
        Minimum size of the assigned and mock groups compared when calculating s_total.
    b0_guess : float
        Initial guess for the b0 parameter.
    r0_guess : float
        Initial guess for the r0 parameter.
    max_stellar_mass : float
        Maximum galactic stellar mass present in the data.

    Returns
    -------
    b_opt : float
        Optimized b0 parameter.
    r_opt : float
        Optimized r0 parameter.
    """

    if redshift_cat.mock_group_ids is None:
        raise InterruptedError(
            "No mock group ids found. Be sure to set the mock groups ids."
        )

    def _objective(params):
        b0, r0 = params
        return -_calc_s_tot(redshift_cat, b0, r0)

    def _calc_s_tot(redshift_cat, b0, r0):
        redshift_cat.run_fof(b0=b0, r0=r0, max_stellar_mass=max_stellar_mass)
        s_tot = redshift_cat.compare_to_mock(min_group_size=min_group_size)
        return s_tot

    res = fmin(
        _objective,
        (b0_guess, r0_guess),
        xtol=0.1,
        ftol=0.1,
        maxiter=50,
        full_output=True,
        disp=False,
    )
    b_opt, r_opt = res[0]

    return b_opt, r_opt



"""
Testing that the group properties are calculated correctly.
"""

import numpy as np
import pandas as pd

from nessie.catalog import RedshiftCatalog
from nessie.cosmology import FlatCosmology
from nessie.helper_funcs import create_density_function


cosmo = FlatCosmology(0.7, 0.3)

SHARK_FILE = (
    "~/Desktop/mock_catalogs/offical_waves_mocks/v0.3.0/wide/waves_wide_gals.parquet"
)

shark_data_frame = pd.read_parquet(SHARK_FILE)
shark_data_frame = shark_data_frame[
    (shark_data_frame["zobs"] < 0.2)
    & (shark_data_frame["total_ab_dust_Z_VISTA"] != -999)
]

redshift = np.array(shark_data_frame["zobs"])
ra = np.array(shark_data_frame["ra"])
dec = np.array(shark_data_frame["dec"])
ab_mag = np.array(shark_data_frame["total_ab_dust_Z_VISTA"])
vel_errs = np.ones(len(shark_data_frame)) * 50


FRAC_AREA = 1133.86 / 41253
func = create_density_function(redshift, len(redshift), FRAC_AREA, cosmo)

red_cat = RedshiftCatalog(
    shark_data_frame["ra"],
    shark_data_frame["dec"],
    shark_data_frame["zobs"],
    func,
    cosmo,
)

red_cat.run_fof(0.05, 18)
group_ids = red_cat.group_ids

group_cat = red_cat.calculate_group_table(ab_mag, vel_errs)


values = {
    "ra": ra,
    "dec": dec,
    "redshift": redshift,
    "absolute_magnitudes": ab_mag,
    "vel_errs": vel_errs,
    "group_ids": group_ids,
}
buggy_values = pd.DataFrame.from_dict(values)
buggy_values.to_csv("test_group_properties.csv")



"""
Testing the core nessie functions in the core_funcs module.
"""

import unittest
import numpy as np

from nessie.core_funcs import _group_graph, _find_groups


class TestGroupGraph(unittest.TestCase):
    """
    Test functions for the _group_graph function.
    """

    def test_two_disconneted_graphs(self):
        """
        Simple case with two groups, one with three members and one with two members.
        """
        links = {"i": np.array([1, 1, 2, 4]), "j": np.array([2, 3, 3, 5])}
        groups = _group_graph(links)

        correct_galaxy_ids = [1, 2, 3, 4, 5]
        correct_group_ids = [1, 1, 1, 2, 2]

        for res, ans in zip(groups["galaxy_id"], correct_galaxy_ids):
            self.assertEqual(res, ans)

        for res, ans in zip(groups["group_id"], correct_group_ids):
            self.assertEqual(res, ans)


class TestFindGroups(unittest.TestCase):
    """
    Test functions for the _find_groups function.
    """

    def test_two_groups(self):
        """
        Simple tests with two groups, 3 members and 2 members and one isolated.
        """
        ra = np.array([120, 120, 120, 0, 0, 180])
        dec = np.array([0, 0, 0, -50, -50, 50])
        comoving_distance = np.array([20, 20, 20, 100, 100, 400])
        los_link_lengths = np.array([1, 1, 1, 1, 1, 1])
        pos_link_lengths = np.array([1, 1, 1, 1, 1, 1])

        correct_galaxy_ids = np.array([1, 2, 3, 4, 5])
        correct_group_ids = np.array([1, 1, 1, 2, 2])

        result = _find_groups(
            ra, dec, comoving_distance, pos_link_lengths, los_link_lengths
        )
        for res, ans in zip(result["galaxy_id"], correct_galaxy_ids):
            self.assertEqual(res, ans)

        for res, ans in zip(result["group_id"], correct_group_ids):
            self.assertEqual(res, ans)


if __name__ == "__main__":
    unittest.main()



"""
Testing the comsology functions in test_core_funcs against astropy and celestial.
"""

import unittest
from astropy.cosmology import FlatLambdaCDM
import numpy as np
import numpy.testing as npt

from nessie.cosmology import FlatCosmology


class TestCoMovingDistance(unittest.TestCase):
    """
    Testing the comoving_distance method.
    """

    cosmo = FlatCosmology(0.7, 0.3)
    astropy_cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

    def test_simple(self):
        """Simple test"""
        redshifts = np.arange(0, 3, 0.1)

        results = self.cosmo.comoving_distance(redshifts)
        answer = self.astropy_cosmo.comoving_distance(redshifts).value
        npt.assert_almost_equal(np.around(results), np.around(answer))


class TestInverseComovingDistance(unittest.TestCase):
    """
    Testing the  z_at_comoving_distances function.
    """

    cosmo = FlatCosmology(0.7, 0.3)

    def test_recovery(self):
        """Check that we get back what we put in."""
        redshifts = np.arange(0, 10, 0.1)
        comoving_distances = self.cosmo.comoving_distance(redshifts)
        returned_redshfits = self.cosmo.z_at_comoving_distances(comoving_distances)
        npt.assert_almost_equal(redshifts, returned_redshfits, decimal=3)


class TestDistanceModulus(unittest.TestCase):
    """
    Testing the dist_mod method in the Cosmology class
    """

    cosmo = FlatCosmology(0.7, 0.3)
    astropy_cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

    def test_astropy(self):
        """
        Simple comparison against astropy.
        """
        redshifts = np.arange(0, 5, 0.1)
        result = self.cosmo.dist_mod(redshifts)
        answer = self.astropy_cosmo.distmod(redshifts).value
        npt.assert_array_almost_equal(result, answer, decimal=3)


class TestVirialRadius(unittest.TestCase):
    """
    Testing the virial_radius method in the Cosmology class.
    """

    cosmo = FlatCosmology(0.7, 0.3)

    def test_against_celestial(self):
        """
        This function is just a pure wrapper for the rust function which is tested there.
        So we need only check that it runs correctly. Testing against the celestial call in R.
        """
        mass = 1e12  # default in celestial
        redshifts = np.array([0.0, 0.1, 1.1])
        answers = np.array([0.2062981, 0.2198784, 0.2859294])
        results = self.cosmo.virial_radius(mass, redshifts)
        npt.assert_almost_equal(results, answers, decimal=3)


class TestVirialSigma(unittest.TestCase):
    """
    Testing the velocity_dispersion method in the Cosmology class.
    """

    cosmo = FlatCosmology(0.7, 0.3)

    def test_against_celestial(self):
        """
        This function is just a pure wrapper for the rust function which is tested there.
        So we need only check that it runs correctly. Testing against the celestial call in R.
        """
        mass = 1e12  # default in celestial
        redshifts = np.array([0.0, 0.1, 1.1])
        answers = np.array([204.2154, 207.4633, 251.3717])
        results = self.cosmo.velocity_dispersion(mass, redshifts)
        npt.assert_almost_equal(results, answers, decimal=1)


class TestHgrow(unittest.TestCase):
    """
    Testing the h0_grow function at
    """

    cosmo = FlatCosmology(0.7, 0.3)
    astropy_cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

    def test_against_astropy(self):
        """
        Tesing the H(z) function against astropy.
        """
        redshifts = np.arange(0, 10, 0.1)
        answers = self.astropy_cosmo.H(redshifts).value
        results = self.cosmo.h0_grow(redshifts)
        npt.assert_almost_equal(results, answers, decimal=2)


if __name__ == "__main__":
    unittest.main()



"""
Module which contains the core functionality for the Nessie group finder.
"""

import networkx as nx
import numpy as np
from nessie_py import fof_links_fast


def _group_graph(links: dict):
    """
    Build a group cataog from a dictionary of links. Involves building the graph and finding
    connected subgraphs.
    """
    graph = nx.Graph()
    edges = zip(links["i"], links["j"])
    graph.add_edges_from(edges)

    components = nx.connected_components(graph)

    group_ids = []
    galaxy_ids = []
    for group_id, component in enumerate(components):
        for galaxy_id in component:
            group_ids.append(group_id)
            galaxy_ids.append(galaxy_id)

    # Mathing the 1-index from R
    group_table = {
        "galaxy_id": np.array(galaxy_ids),
        "group_id": np.array(group_ids) + 1,
    }
    return group_table


def _find_groups(
    ra: np.ndarray[float],
    dec: np.ndarray[float],
    comoving_distance: np.ndarray[float],
    linking_length_pos: np.ndarray[float],
    linking_length_los: np.ndarray[float],
) -> dict:
    """
    Identify groups within the given ra, dec, and comoving distance arrays, given the los and pos
    linking lengths.
    """
    links = fof_links_fast(
        ra, dec, comoving_distance, linking_length_pos, linking_length_los
    )
    return _group_graph(links)



"""
Simple example script. Testing benchmarks, scores, and other integrations.
"""

import numpy as np
import pandas as pd

from nessie.catalog import RedshiftCatalog
from nessie.cosmology import FlatCosmology
from nessie.helper_funcs import create_density_function, calculate_s_total

cosmo = FlatCosmology(0.7, 0.3)


random_zs = np.loadtxt(
    "/Users/00115372/Desktop/GAMA_paper_plotter/gama_g09_randoms.txt", skiprows=1
)
func = create_density_function(random_zs, len(random_zs) / 400, 0.001453924, cosmo)

data = pd.read_parquet("~/Desktop/GAMA_paper_plotter/mocks/galform_gals_for_R.parquet")
data = data[data["Volume"] == 1]
print("data read in")

ra = np.array(data["RA"])
dec = np.array(data["DEC"])
redshift = np.array(data["Zspec"])

red_cat = RedshiftCatalog(ra, dec, redshift, func, cosmo)
red_cat.run_fof(b0=0.05, r0=18)

red_cat.mock_group_ids = np.array(data["GroupID"])
red_cat.completeness = np.ones(len(ra)) * 0.95

score = red_cat.compare_to_mock(min_group_size=5)
print(score)

another_score = calculate_s_total(red_cat.group_ids, red_cat.mock_group_ids, 5)
print(another_score)



"""
Integration test testing the optimizer function
"""

from datetime import datetime

import numpy as np

from nessie.catalog import RedshiftCatalog
from nessie.cosmology import FlatCosmology
from nessie.optimizer import optimize_nm
from nessie.helper_funcs import create_density_function


INFILE_SDSS = "/Users/00115372/Desktop/nessie_plots/asu.tsv"
SDSS_AREA = 0.212673
group_id, n_gal, z, ra, dec, mag = np.loadtxt(INFILE_SDSS, unpack=True, skiprows=1)
group_id[group_id == 0] = -1

cosmo = FlatCosmology(0.7, 0.3)

func = create_density_function(z, len(z), SDSS_AREA, cosmo)

red_cat = RedshiftCatalog(ra, dec, z, func, cosmo)
red_cat.set_completeness()
red_cat.mock_group_ids = group_id


start = datetime.now()
b0, r0 = optimize_nm(red_cat, 5)
end = datetime.now()
print("Time taken to optimize is: ", end - start)
